/*
 *  drivers/s390/char/tape_3590.c
 *    tape device discipline for 3590 tapes.
 *
 *    Copyright (C) IBM Corp. 2001,2006
 *    Author(s): Stefan Bader <shbader@de.ibm.com>
 *               Michael Holzheu <holzheu@de.ibm.com>
 *               Martin Schwidefsky <schwidefsky@de.ibm.com>
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <asm/tape390.h>

#define TAPE_DBF_AREA	tape_3590_dbf

#include "tape.h"
#include "tape_std.h"
#include "tape_3590.h"

/*
 * Pointer to debug area.
 */
debug_info_t *TAPE_DBF_AREA = NULL;

/*******************************************************************
 * Error Recovery fuctions:
 * - Read Opposite:		 implemented
 * - Read Device (buffered) log: BRA
 * - Read Library log:		 BRA
 * - Swap Devices:		 BRA
 * - Long Busy:			 BRA
 * - Special Intercept:		 BRA
 * - Read Alternate:		 implemented
 *******************************************************************/

#define PRINTK_HEADER "T3590:"

static const char* tape_3590_msg[TAPE_3590_MAX_MSG] = {
	[0x00] = "",
	[0x10] = "Lost Sense",
	[0x11] = "Assigned Elsewhere",
	[0x12] = "Allegiance Reset",
	[0x13] = "Shared Access Violation",
	[0x20] = "Command Reject",
	[0x21] = "Configuration Error",
	[0x22] = "Protection Exception",
	[0x23] = "Write Protect",
	[0x24] = "Write Length",
	[0x25] = "Read-Only Format",
	[0x31] = "Beginning of Partition",
	[0x33] = "End of Partition",
	[0x34] = "End of Data",
	[0x35] = "Block not found",
	[0x40] = "Device Intervention",
	[0x41] = "Loader Intervention",
	[0x42] = "Library Intervention",
	[0x50] = "Write Error",
	[0x51] = "Erase Error",
	[0x52] = "Formatting Error",
	[0x53] = "Read Error",
	[0x54] = "Unsupported Format",
	[0x55] = "No Formatting",
	[0x56] = "Positioning lost",
	[0x57] = "Read Length",
	[0x60] = "Unsupported Medium",
	[0x61] = "Medium Length Error",
	[0x62] = "Medium removed",
	[0x64] = "Load Check",
	[0x65] = "Unload Check",
	[0x70] = "Equipment Check",
	[0x71] = "Bus out Check",
	[0x72] = "Protocol Error",
	[0x73] = "Interface Error",
	[0x74] = "Overrun",
	[0x75] = "Halt Signal",
	[0x90] = "Device fenced",
	[0x91] = "Device Path fenced",
	[0xa0] = "Volume misplaced",
	[0xa1] = "Volume inaccessible",
	[0xa2] = "Volume in input",
	[0xa3] = "Volume ejected",
	[0xa4] = "All categories reserved",
	[0xa5] = "Duplicate Volume",
	[0xa6] = "Library Manager Offline",
	[0xa7] = "Library Output Station full",
	[0xa8] = "Vision System non-operational",
	[0xa9] = "Library Manager Equipment Check",
	[0xaa] = "Library Equipment Check",
	[0xab] = "All Library Cells full",
	[0xac] = "No Cleaner Volumes in Library",
	[0xad] = "I/O Station door open",
	[0xae] = "Subsystem environmental alert",
};

/*******************************************************************
 * Private IOCTLS
 *******************************************************************/

/*
 * Read config data
 */
static int
tape_3590_read_config (struct tape_device *device,
		       struct tape_3590_diag_data *diag)
{
	static int dcp0123[4] = { 0x00000000, 0x01000000,
				  0x02000000, 0x03000000 };
	struct tape_request *request;

	request = tape_alloc_request(9, 16);
	if (IS_ERR(request))
		return PTR_ERR(request);
	request->op = TO_READ_CONFIG;
	/* setup ccws */
	memcpy(request->cpdata, dcp0123, sizeof(dcp0123));
	tape_ccw_cc(request->cpaddr, MODE_SENSE, 4, request->cpdata);
	tape_ccw_cc(request->cpaddr + 1, READ_SS_DATA, 36, diag->dcp0);
	tape_ccw_cc(request->cpaddr + 2, MODE_SENSE, 4, request->cpdata + 4);
	tape_ccw_cc(request->cpaddr + 3, READ_SS_DATA, 36, diag->dcp1);
	tape_ccw_cc(request->cpaddr + 4, MODE_SENSE, 4, request->cpdata + 8);
	tape_ccw_cc(request->cpaddr + 5, READ_SS_DATA, 36, diag->dcp2);
	tape_ccw_cc(request->cpaddr + 6, MODE_SENSE, 4, request->cpdata + 12);
	tape_ccw_cc(request->cpaddr + 7, READ_SS_DATA, 36, diag->dcp3);
	tape_ccw_end(request->cpaddr + 8, NOP, 0, NULL);
	/* execute it */
	return tape_do_io_free(device, request);
}

/*
 * Diagnose IOCTL
 */
static int
tape_3590_ioctl_diag(struct tape_device *device, unsigned long arg)
{
	struct tape_3590_diag_data *diag;
	int rc;
	
	if (!capable(CAP_SYS_ADMIN))
		return -EACCES;
	diag = kmalloc(sizeof(struct tape_3590_diag_data),
		       GFP_KERNEL | GFP_DMA);
	if (!diag)
		return -ENOMEM;
	rc = tape_3590_read_config(device, diag);
	if (rc == 0) {
		if (copy_to_user ((char *) arg, diag,
				  sizeof(struct tape_3590_diag_data)))
			rc = -EFAULT;
	}
	kfree(diag);
	return rc;
}

/*
 * 3590 IOCTL Overload
 */
static int
tape_3590_ioctl(struct tape_device *device,
		unsigned int cmd, unsigned long arg)
{
	switch (cmd) {
	case TAPE_3590_IOCTL_DIAG:
		return tape_3590_ioctl_diag(device, arg);
	case TAPE390_DISPLAY: {
		struct display_struct disp;

		if(copy_from_user(&disp, (char *) arg, sizeof(disp)) != 0)
			return -EFAULT;

		return tape_std_display(device, &disp);
	}
	default:
		return -EINVAL;		// no additional ioctls
	}
}


/*
 * SENSE Medium: Get Sense data about medium state
 */
static void
__tape_3590_sense_medium(struct tape_request *request, void *data)
{
	struct tape_3590_med_sense *sense;

	request->callback = NULL;

	if (request->rc < 0) {
		tape_med_state_set(request->device, MS_UNKNOWN);
	} else {
		sense = (struct tape_3590_med_sense *) request->cpdata;
		if (sense->masst == MSENSE_UNASSOCIATED)
			tape_med_state_set(request->device, MS_UNLOADED);
		if (sense->masst == MSENSE_ASSOCIATED_MOUNT)
			tape_med_state_set(request->device, MS_LOADED);
	}

	tape_put_request(request);
}

static int
tape_3590_sense_medium (struct tape_device  *device)
{
	struct tape_request *request;

	request = tape_alloc_request(1, 128);
	if (IS_ERR(request))
		return PTR_ERR(request);

	request->op = TO_MSEN;
	request->callback = __tape_3590_sense_medium;

	/* setup ccws */
	tape_ccw_end(request->cpaddr, MEDIUM_SENSE, 128, request->cpdata);

	/* execute it */
	return tape_do_io_async(device, request);
}

static void
tape_3590_work_handler(void *data)
{
	struct {
		struct tape_device	*device;
		enum tape_op		 op;
		struct tq_struct	 task;
	} *p = data;

	switch(p->op) {
		case TO_MSEN:
			tape_3590_sense_medium(p->device);
			break;
		default:
			DBF_EVENT(3, "T3590: internal error: unknown work\n");
	}

	tape_put_device(p->device);
	kfree(p);
}

static int
tape_3590_schedule_work(struct tape_device *device, enum tape_op op)
{
	struct {
		struct tape_device	*device;
		enum tape_op		 op;
		struct tq_struct	 task;
	} *p;

	if ((p = kmalloc(sizeof(*p), GFP_ATOMIC)) == NULL)
		return -ENOMEM;

	memset(p, 0, sizeof(*p));
	INIT_LIST_HEAD(&p->task.list);
	p->task.routine = tape_3590_work_handler;
	p->task.data    = p;

	p->device = tape_clone_device(device);
	p->op     = op;

	schedule_task(&p->task);

	return 0;
}

/*
 * MTTELL: Tell block. Return the number of block relative to current file.
 */
int
tape_3590_mttell(struct tape_device *device, int mt_count)
{
	unsigned int	bid;
	int		rc;

	rc = tape_std_read_block_id(device, &bid);
	if (rc)
		return rc;

	return bid;
}

/*
 * MTSEEK: seek to the specified block.
 */
static int
tape_3590_mtseek (struct tape_device  *device, int count)
{
	struct tape_request *request;

	DBF_EVENT(6, "xsee id: %x\n", count);
	request = tape_alloc_request(3, 4);
	if (IS_ERR(request))
		return PTR_ERR(request);
	request->op = TO_LBL;
	/* setup ccws */
	tape_ccw_cc(request->cpaddr, MODE_SET_DB, 1, device->modeset_byte);
	*(__u32 *) request->cpdata = count;
	tape_ccw_cc(request->cpaddr + 1, LOCATE, 4, request->cpdata);
	tape_ccw_end(request->cpaddr + 2, NOP, 0, NULL);
	/* execute it */
	return tape_do_io_free(device, request);
}

/*
 * Read Opposite Error Recovery Function:
 * Used, when Read Forward does not work
 */
static void
tape_3590_read_opposite (struct tape_device *device,
			 struct tape_request *request)
{
	struct tape_3590_disc_data *data;
	
	/*
	 * We have allocated 4 ccws in tape_std_read, so we can now
	 * transform the request to a read backward, followed by a
	 * forward space block.
	 */
	request->op = TO_RBA;
	tape_ccw_cc(request->cpaddr, MODE_SET_DB, 1, device->modeset_byte);
	data = (struct tape_3590_disc_data *) device->discdata;
	tape_ccw_cc_idal(request->cpaddr + 1, data->read_back_op,
			 device->char_data.idal_buf);
	tape_ccw_cc(request->cpaddr + 2, FORSPACEBLOCK, 0, NULL);
	tape_ccw_end(request->cpaddr + 3, NOP, 0, NULL);
	DBF_EVENT(6, "xrop ccwg\n");
}

#ifdef CONFIG_S390_TAPE_BLOCK
/*
 * Tape Block READ
 */
struct tape_request *
tape_3590_bread (struct tape_device *device, struct request *req)
{
	struct tape_request *request;
	struct buffer_head  *bh;
	ccw1_t              *ccw;
	int                  count;
	int                  size;
	int                  start_block;

	DBF_EVENT(6, "xBREDid:");
	start_block = req->sector >> TAPEBLOCK_HSEC_S2B;
	DBF_EVENT(6, "start_block = %i\n", start_block);

	count = 0;
	for(bh = req->bh; bh; bh = bh->b_reqnext) {
		for(size = 0; size < bh->b_size; size += TAPEBLOCK_HSEC_SIZE)
			count++;
	}

	request = tape_alloc_request(2+count+1, 4);
	if (IS_ERR(request))
		return request;

	request->op = TO_BLOCK;
	*(__u32 *) request->cpdata = start_block;
	ccw = request->cpaddr;
	ccw = tape_ccw_cc(ccw, MODE_SET_DB, 1, device->modeset_byte);

	/*
	 * We always setup a nop after the mode set ccw. This slot is
	 * used in tape_std_check_locate to insert a locate ccw if the
	 * current tape position doesn't match the start block to be read.
	 */
	ccw = tape_ccw_cc(ccw, NOP, 0, NULL);

	for(bh = req->bh; bh; bh = bh->b_reqnext) {
		for(size = 0; size < bh->b_size; size += TAPEBLOCK_HSEC_SIZE) {
			ccw->flags = CCW_FLAG_CC;
			ccw->cmd_code = READ_FORWARD;
			ccw->count = TAPEBLOCK_HSEC_SIZE;
			set_normalized_cda(ccw, (void *) __pa(bh->b_data+size));
			ccw++;
		}
	}

	ccw = tape_ccw_end(ccw, NOP, 0, NULL);
	DBF_EVENT(6, "xBREDccwg\n");

	return request;
}

void
tape_3590_free_bread (struct tape_request *request)
{
	ccw1_t* ccw;

	for (ccw = request->cpaddr; ccw->flags & CCW_FLAG_CC; ccw++)
		if (ccw->cmd_code == READ_FORWARD)
			clear_normalized_cda(ccw);
	tape_put_request(request);
}

/*
 * check_locate is called just before the tape request is passed to
 * the common io layer for execution. It has to check the current
 * tape position and insert a locate ccw if it doesn't match the
 * start block for the request.
 */
void
tape_3590_check_locate (struct tape_device *device,
			struct tape_request *request)
{
	int *start_block = (__u32 *) request->cpdata;

	if (*start_block != device->blk_data.block_position) {
		*start_block += device->blk_data.start_block_id;
		tape_ccw_cc(request->cpaddr + 1, LOCATE, 4, request->cpdata);
	}
}
#endif

/*
 * The done handler is called at device/channel end and wakes up the sleeping
 * process
 */
static int
tape_3590_done(struct tape_device *device, struct tape_request *request)
{
	struct tape_3590_med_sense *sense;

	DBF_EVENT(6,"%s done\n", tape_op_verbose[request->op]);

	// FIXME: Maybe only on assign/unassign
	TAPE_CLEAR_STATE(device, TAPE_STATUS_BOXED);

	switch(request->op) {
	case TO_BSB:
	case TO_BSF:
	case TO_DSE:
	case TO_FSB:
	case TO_FSF:
	case TO_LBL:
	case TO_RFO:
	case TO_RBA:
	case TO_REW:
	case TO_WRI:
	case TO_WTM:
	case TO_BLOCK:
	case TO_LOAD:
		tape_med_state_set(device,MS_LOADED);
		break;
	case TO_RUN:
		tape_med_state_set(device,MS_UNLOADED);
		break;
	case TO_MSEN:
		sense = (struct tape_3590_med_sense *) request->cpdata;
		if (sense->masst == MSENSE_UNASSOCIATED)
			tape_med_state_set(device,MS_UNLOADED);
		if (sense->masst == MSENSE_ASSOCIATED_MOUNT)
			tape_med_state_set(device,MS_LOADED);
		break;
	case TO_RBI: /* RBI seems to succeed even without medium loaded. */
	case TO_NOP: /* Same to NOP. */
	case TO_READ_CONFIG:
	case TO_READ_ATTMSG:
	case TO_DIS:
	case TO_ASSIGN:
	case TO_UNASSIGN:
	case TO_BREAKASS:
		break;
	case TO_SIZE:
		break;
	}
	return TAPE_IO_SUCCESS;
}

/*
 * This fuction is called, when error recovery was successful
 */
static inline int
tape_3590_erp_succeded(struct tape_device *device,
		       struct tape_request *request)
{
	DBF_EVENT(3, "Error Recovery successful for %s\n",
		  tape_op_verbose[request->op]);
	return tape_3590_done(device, request);
}

/*
 * This fuction is called, when error recovery was not successful
 */
static inline int
tape_3590_erp_failed(struct tape_device *device,
		     struct tape_request *request, int rc)
{
	DBF_EVENT(3, "Error Recovery failed for %s\n",
		  tape_op_verbose[request->op]);
	tape_dump_sense_dbf(device, request);
	return rc;
}

/*
 * Error Recovery do retry
 */
static inline int
tape_3590_erp_retry(struct tape_device *device, struct tape_request *request)
{
	DBF_EVENT(2, "Retry: %s\n", tape_op_verbose[request->op]);
	tape_dump_sense_dbf(device, request);
	return TAPE_IO_RETRY;
}

/*
 * Handle unsolicited interrupts
 */
static int
tape_3590_unsolicited_irq(struct tape_device *device)
{
	if (device->devstat.dstat == 0x85)
		/* A medium was inserted in the driver. */
		tape_3590_schedule_work(device, TO_MSEN);
	else if (device->devstat.dstat == DEV_STAT_CHN_END)
		/* Probably result of halt ssch */
		return TAPE_IO_PENDING;
	else if (device->devstat.dstat & DEV_STAT_ATTENTION) {
		DBF_EVENT(2, "Unsolicited unit attention (flag=%d)!\n",
			device->devstat.flag);

		/*
		 * Currently we don't want to read the attention message.
		 * (see read_attmsg)
		 * tape_3590_schedule_work(device, TO_READ_ATTMSG);
		 */
	} else {
		DBF_EVENT(3, "unsol.irq! dev end: %x\n",
			  device->devinfo.irq);
		PRINT_WARN("Unsolicited IRQ (Device End) caught.\n");
		tape_dump_sense(device, NULL);
	}
	return TAPE_IO_SUCCESS;
}

/*
 * Basic Recovery routine which handles Recovery Action 0xc0
 */
static int
tape_3590_erp_basic(struct tape_device *device,
		    struct tape_request *request, int rc)
{
	struct tape_3590_sense *sense;

	sense = (struct tape_3590_sense *) device->devstat.ii.sense.data;

	switch (sense->bra) {
	case SENSE_BRA_PER:
		return tape_3590_erp_failed(device, request, rc);
	case SENSE_BRA_CONT:
		return tape_3590_erp_succeded(device, request);
	case SENSE_BRA_RE:
		return tape_3590_erp_retry(device, request);
	case SENSE_BRA_DRE:
		return tape_3590_erp_failed(device, request, rc);
	default:
		PRINT_ERR("Unknown BRA %x - This should not happen!\n",
			  sense->bra);
		BUG();
		return TAPE_IO_STOP;
	}
}

/*
 *  RDL: Read Device (buffered) log
 */
static int
tape_3590_erp_read_buf_log(struct tape_device *device,
			   struct tape_request *request)
{
	/*
	 * we get a read bufferd log unit ckeck, e.g. if
	 * one performs an unload. We just do the basic
	 * error recovery at the moment (retry). Perhaps
	 * in the  future, we read the log and dump it
	 * somewhere...
	 */
	return tape_3590_erp_basic(device, request, -EIO);
}

/*
 *  SWAP: Swap Devices
 */
static int
tape_3590_erp_swap (struct tape_device *device, struct tape_request *request)
{
	/*
	 * This error recovery should swap the tapes
	 * if the original has a problem. The operation
	 * should proceed with the new tape... this
	 * should probably be done in user space!
	 */
	PRINT_WARN("(%x): Swap Tape Device!\n", device->devstat.devno);
	return tape_3590_erp_basic(device, request, -EIO);
}

/*
 *  LBY: Long Busy
 */
static int
tape_3590_erp_long_busy(struct tape_device *device,
			struct tape_request *request)
{
	PRINT_WARN("(%x): Device is busy! Please wait a minute!\n",
		   device->devstat.devno);
	return tape_3590_erp_basic(device, request, -EBUSY);
}

/*
 *  SPI: Special Intercept
 */
static int
tape_3590_erp_special_interrupt(struct tape_device *device,
				struct tape_request *request)
{
	return tape_3590_erp_basic(device, request, -EIO);
}

/*
 *  RDA: Read Alternate
 */
static int
tape_3590_erp_read_alternate(struct tape_device *device,
			     struct tape_request *request)
{
	struct tape_3590_disc_data *data;

	/*
	 * The issued Read Backward or Read Previous command is not
	 * supported by the device
	 * The recovery action should be to issue another command:
	 * Read Revious: if Read Backward is not supported
	 * Read Backward: if Read Previous is not supported
	 */
	data = (struct tape_3590_disc_data *) device->discdata;
	if (data->read_back_op == READ_PREVIOUS) {
		DBF_EVENT(2, "(%x): No support for READ_PREVIOUS command\n",
			  device->devstat.devno);
		data->read_back_op = READ_BACKWARD;
	} else {
		DBF_EVENT(2, "(%x): No support for READ_BACKWARD command\n",
			  device->devstat.devno);
		data->read_back_op = READ_PREVIOUS;
	}
	tape_3590_read_opposite(device, request);
	return tape_3590_erp_retry(device, request);
}

/*
 * Error Recovery read opposite
 */
static int
tape_3590_erp_read_opposite(struct tape_device *device,
			    struct tape_request *request)
{
	switch (request->op) {
	case TO_RFO:
		/*
		 * We did read forward, but the data could not be read.
		 * We will read backward and then skip forward again.
		 */
		tape_3590_read_opposite(device, request);
		return tape_3590_erp_retry(device, request);
	case TO_RBA:
		/* We tried to read forward and backward, but hat no success */
		return tape_3590_erp_failed(device, request, -EIO);
		break;
	default:
		PRINT_WARN("read_opposite_recovery_called_with_op: %s\n",
			   tape_op_verbose[request->op]);
		return tape_3590_erp_failed(device, request, -EIO);
	}
}

/*
 * Print an MIM (Media Information  Message)
 */
static void
tape_3590_print_mim_msg_f0(struct tape_device *device)
{
	struct tape_3590_sense *sense;

	sense = (struct tape_3590_sense *) device->devstat.ii.sense.data;
	/* Exception Message */
	switch(sense->fmt.f70.emc) {
	case 0x02:
		PRINT_WARN("(%x): Data degraded\n", device->devstat.devno);
		break;
	case 0x03:
		PRINT_WARN("(%x): Data degraded in partion %i\n",
			   device->devstat.devno, sense->fmt.f70.mp);
		break;
	case 0x04:
		PRINT_WARN("(%x): Medium degraded\n", device->devstat.devno);
		break;
	case 0x05:
		PRINT_WARN("(%x): Medium degraded in partition %i\n",
			   device->devstat.devno, sense->fmt.f70.mp);
		break;
	case 0x06:
		PRINT_WARN("(%x): Block 0 Error\n", device->devstat.devno);
		break;
	case 0x07:
		PRINT_WARN("(%x): Medium Exception 0x%02x\n",
			   device->devstat.devno,  sense->fmt.f70.md);
		break;
	default:
		PRINT_WARN("(%x): MIM ExMsg: 0x%02x\n",
			   device->devstat.devno, sense->fmt.f70.emc);
		break;
	}
	/* Service Message */
	switch(sense->fmt.f70.smc) {
	case 0x02:
		PRINT_WARN("(%x): Reference Media maintenance procedure %i\n",
			   device->devstat.devno, sense->fmt.f70.md);
		break;
	default:
		PRINT_WARN("(%x): MIM ServiceMsg: 0x%02x\n",
			   device->devstat.devno, sense->fmt.f70.smc);
		break;
	}
}

/*
 * Print an I/O Subsystem Service Information Message
 */
static void
tape_3590_print_io_sim_msg_f1(struct tape_device *device)
{
	struct tape_3590_sense *sense;

	sense = (struct tape_3590_sense *) device->devstat.ii.sense.data;
	/* Exception Message */
	switch(sense->fmt.f71.emc) {
	case 0x01:
		PRINT_WARN("(%x): Effect of failure is unknown\n",
			   device->devstat.devno);
		break;
	case 0x02:
		PRINT_WARN("(%x): CU Exception - no performance impact\n",
			   device->devstat.devno);
		break;
	case 0x03:
		PRINT_WARN("(%x): CU Exception on channel interface 0x%02x\n",
			   device->devstat.devno, sense->fmt.f71.md[0]);
		break;
	case 0x04:
		PRINT_WARN("(%x): CU Exception on device path 0x%02x\n",
			   device->devstat.devno, sense->fmt.f71.md[0]);
		break;
	case 0x05:
		PRINT_WARN("(%x): CU Exception on library path 0x%02x\n",
			   device->devstat.devno, sense->fmt.f71.md[0]);
		break;
	case 0x06:
		PRINT_WARN("(%x): CU Exception on node 0x%02x\n",
			   device->devstat.devno, sense->fmt.f71.md[0]);
		break;
	case 0x07:
		PRINT_WARN("(%x): CU Exception on partition 0x%02x\n",
			   device->devstat.devno, sense->fmt.f71.md[0]);
		break;
	default:
		PRINT_WARN("(%x): SIM ExMsg: 0x%02x\n",
			   device->devstat.devno, sense->fmt.f71.emc);
	}
	/* Service Message */
	switch(sense->fmt.f71.smc) {
	case 0x01:
		PRINT_WARN("(%x): Repair impact is unknown\n",
			   device->devstat.devno);
		break;
	case 0x02:
		PRINT_WARN("(%x): Repair will not impact cu performance\n",
			   device->devstat.devno);
		break;
	case 0x03:
		if (sense->fmt.f71.mdf == 0)
			PRINT_WARN("(%x): Repair will disable node "
				   "0x%x on CU\n",
				   device->devstat.devno,
				   sense->fmt.f71.md[1]);
		else
			PRINT_WARN("(%x): Repair will disable nodes "
				   "(0x%x-0x%x) on CU\n",
				   device->devstat.devno,
				   sense->fmt.f71.md[1],
				   sense->fmt.f71.md[2]);
		break;
	case 0x04:
		if (sense->fmt.f71.mdf == 0)
			PRINT_WARN("(%x): Repair will disable cannel path "
				   "0x%x on CU\n",
				   device->devstat.devno,
				   sense->fmt.f71.md[1]);
		else
			PRINT_WARN("(%x): Repair will disable cannel paths "
				   "(0x%x-0x%x) on CU\n",
				   device->devstat.devno,
				   sense->fmt.f71.md[1],
				   sense->fmt.f71.md[2]);
		break;
	case 0x05:
		if (sense->fmt.f71.mdf == 0)
			PRINT_WARN("(%x): Repair will disable device path "
				   "0x%x on CU\n",
				   device->devstat.devno,
				   sense->fmt.f71.md[1]);
		else
			PRINT_WARN("(%x): Repair will disable device paths "
				   "(0x%x-0x%x) on CU\n",
				   device->devstat.devno,
				   sense->fmt.f71.md[1],
				   sense->fmt.f71.md[2]);
		break;
	case 0x06:
		if (sense->fmt.f71.mdf == 0)
			PRINT_WARN("(%x): Repair will disable library path "
				   "0x%x on CU\n",
				   device->devstat.devno,
				   sense->fmt.f71.md[1]);
		else
			PRINT_WARN("(%x): Repair will disable library paths "
				   "(0x%x-0x%x) on CU\n",
				   device->devstat.devno,
				   sense->fmt.f71.md[1],
				   sense->fmt.f71.md[2]);
		break;
	case 0x07:
		PRINT_WARN("(%x): Repair will disable access to CU\n",
			   device->devstat.devno);
		break;
	default:
		PRINT_WARN("(%x): SIM ServiceMsg: 0x%02x\n",
			   device->devstat.devno, sense->fmt.f71.smc);
	}
}

/*
 * Print an Device Subsystem Service Information Message
 */
static void
tape_3590_print_dev_sim_msg_f2(struct tape_device *device)
{
	struct tape_3590_sense *sense;
	
	sense = (struct tape_3590_sense *) device->devstat.ii.sense.data;
	/* Exception Message */
	switch(sense->fmt.f71.emc) {
	case 0x01:
		PRINT_WARN("(%x): Effect of failure is unknown\n",
			   device->devstat.devno);
		break;
	case 0x02:
		PRINT_WARN("(%x): DV Exception - no performance impact\n",
			   device->devstat.devno);
		break;
	case 0x03:
		PRINT_WARN("(%x): DV Exception on channel interface 0x%02x\n",
			   device->devstat.devno, sense->fmt.f71.md[0]);
		break;
	case 0x04:
		PRINT_WARN("(%x): DV Exception on loader 0x%02x\n",
			   device->devstat.devno, sense->fmt.f71.md[0]);
		break;
	case 0x05:
		PRINT_WARN("(%x): DV Exception on message display 0x%02x\n",
			   device->devstat.devno, sense->fmt.f71.md[0]);
		break;
	case 0x06:
		PRINT_WARN("(%x): DV Exception in tape path\n",
			   device->devstat.devno);
		break;
	case 0x07:
		PRINT_WARN("(%x): DV Exception in drive\n",
			   device->devstat.devno);
		break;
	default:
		PRINT_WARN("(%x): DSIM ExMsg: 0x%02x\n",
			   device->devstat.devno, sense->fmt.f71.emc);
	}
	/* Service Message */
	switch(sense->fmt.f71.smc) {
	case 0x01:
		PRINT_WARN("(%x): Repair impact is unknown\n",
			   device->devstat.devno);
		break;
	case 0x02:
		PRINT_WARN("(%x): Repair will not impact device performance\n",
			   device->devstat.devno);
		break;
	case 0x03:
		if (sense->fmt.f71.mdf == 0)
			PRINT_WARN("(%x): Repair will disable channel path "
				   "0x%x on DV\n",
				   device->devstat.devno,
				   sense->fmt.f71.md[1]);
		else
			PRINT_WARN("(%x): Repair will disable channel path "
				   "(0x%x-0x%x) on DV\n",
				   device->devstat.devno,
				   sense->fmt.f71.md[1],
				   sense->fmt.f71.md[2]);
		break;
	case 0x04:
		if (sense->fmt.f71.mdf == 0)
			PRINT_WARN("(%x): Repair will disable interface 0x%x "
				   "on DV\n",
				   device->devstat.devno,
				   sense->fmt.f71.md[1]);
		else
			PRINT_WARN("(%x): Repair will disable interfaces "
				   "(0x%x-0x%x) on DV\n",
				   device->devstat.devno,
				   sense->fmt.f71.md[1],
				   sense->fmt.f71.md[2]);
		break;
	case 0x05:
		if (sense->fmt.f71.mdf == 0)
			PRINT_WARN("(%x): Repair will disable loader 0x%x "
				   "on DV\n",
				   device->devstat.devno,
				   sense->fmt.f71.md[1]);
		else
			PRINT_WARN("(%x): Repair will disable loader "
				   "(0x%x-0x%x) on DV\n",
				   device->devstat.devno,
				   sense->fmt.f71.md[1],
				   sense->fmt.f71.md[2]);
		break;
	case 0x07:
		PRINT_WARN("(%x): Repair will disable access to DV\n",
			   device->devstat.devno);
		break;
	case 0x08:
		if (sense->fmt.f71.mdf == 0)
			PRINT_WARN("(%x): Repair will disable message "
				   "display 0x%x on DV\n",
				   device->devstat.devno,
				   sense->fmt.f71.md[1]);
		else
			PRINT_WARN("(%x): Repair will disable message "
				   "displays (0x%x-0x%x) on DV\n",
				   device->devstat.devno,
				   sense->fmt.f71.md[1],
				   sense->fmt.f71.md[2]);
		break;
	case 0x09:
		PRINT_WARN("(%x): Clean DV\n", device->devstat.devno);
		break;
	default:
		PRINT_WARN("(%x): DSIM ServiceMsg: 0x%02x\n",
			   device->devstat.devno, sense->fmt.f71.smc);
	}
}

/*
 * Print standard ERA Message
 */
static void
tape_3590_print_era_msg(struct tape_device *device)
{
	struct tape_3590_sense *sense;

	sense = (struct tape_3590_sense *) device->devstat.ii.sense.data;
	if (sense->mc == 0)
		return;
	if ((sense->mc > 0) && (sense->mc < TAPE_3590_MAX_MSG)) {
		if (tape_3590_msg[sense->mc] != NULL)
			PRINT_WARN("(%x): %s\n", device->devstat.devno,
				   tape_3590_msg[sense->mc]);
		else{
			PRINT_WARN("(%x): Message Code 0x%x\n",
				   device->devstat.devno,sense->mc);
		}
		return;
	}
	if (sense->mc == 0xf0) {
		/* Standard Media Information Message */
		PRINT_WARN("(%x): MIM SEV=%i, MC=%02x, ES=%x/%x, "
			   "RC=%02x-%04x-%02x\n", device->devstat.devno,
			   sense->fmt.f70.sev, sense->mc,
			   sense->fmt.f70.emc, sense->fmt.f70.smc,
			   sense->fmt.f70.refcode, sense->fmt.f70.mid,
			   sense->fmt.f70.fid);
		tape_3590_print_mim_msg_f0(device);
		return;
	}
	if (sense->mc == 0xf1) {
		/* Standard I/O Subsystem Service Information Message */
		PRINT_WARN("(%x): IOSIM SEV=%i, DEVTYPE=%04x/%02x, "
			   "MC=%02x, ES=%x/%x, REF=0x%04x-0x%04x-0x%04x\n",
			   device->devstat.devno, sense->fmt.f71.sev,
			   device->devinfo.sid_data.dev_type,
			   device->devinfo.sid_data.dev_model,
			   sense->mc, sense->fmt.f71.emc,
			   sense->fmt.f71.smc, sense->fmt.f71.refcode1,
			   sense->fmt.f71.refcode2, sense->fmt.f71.refcode3);
		tape_3590_print_io_sim_msg_f1(device);
		return;
	}
	if (sense->mc == 0xf2) {
		/* Standard Device Service Information Message */
		PRINT_WARN("(%x): DEVSIM SEV=%i, DEVTYPE=%04x/%02x, "
			   "MC=%02x, ES=%x/%x, REF=0x%04x-0x%04x-0x%04x\n",
			   device->devstat.devno, sense->fmt.f71.sev,
			   device->devinfo.sid_data.dev_type,
			   device->devinfo.sid_data.dev_model,
			   sense->mc, sense->fmt.f71.emc,
			   sense->fmt.f71.smc, sense->fmt.f71.refcode1,
			   sense->fmt.f71.refcode2, sense->fmt.f71.refcode3);
		tape_3590_print_dev_sim_msg_f2(device);
		return;
	}
	if (sense->mc == 0xf3) {
		/* Standard Library Service Information Message */
		return;
	}
	PRINT_WARN("(%x): Device Message(%x)\n",
		   device->devstat.devno, sense->mc);
}

/*
 *  3590 error Recovery routine:
 *  If possible, it tries to recover from the error. If this is not possible,
 *  inform the user about the problem.
 */
static int
tape_3590_unit_check(struct tape_device *device, struct tape_request *request)
{
	struct tape_3590_sense *sense;
	int rc;

#ifdef CONFIG_S390_TAPE_BLOCK
	if (request->op == TO_BLOCK) {
		/*
		 * Recovery for block device requests. Set the block_position
		 * to something invalid and retry.
		 */
		device->blk_data.block_position = -1;
		if (request->retries-- <= 0)
			return tape_3590_erp_failed(device, request, -EIO);
		else
			return tape_3590_erp_retry(device, request);
	}
#endif

	sense = (struct tape_3590_sense *) device->devstat.ii.sense.data;

	/*
	 * First check all RC-QRCs where we want to do something special
	 *   - "break":	    basic error recovery is done
	 *   - "goto out:": just print error message if available
	 */
	rc = -EIO;
	switch(sense->rc_rqc) {
	case 0x1110:
		tape_3590_print_era_msg(device);
		return tape_3590_erp_read_buf_log(device, request);
	case 0x2011:
		tape_3590_print_era_msg(device);
		return tape_3590_erp_read_alternate(device, request);
	case 0x2230:
	case 0x2231:
		tape_3590_print_era_msg(device);
		return tape_3590_erp_special_interrupt(device, request);
	case 0x3010:
		DBF_EVENT(2, "(%x): Backward at Beginning of Partition\n",
			  device->devstat.devno);
		return tape_3590_erp_basic(device, request, -ENOSPC);
	case 0x3012:
		DBF_EVENT(2, "(%x): Forward at End of Partition\n",
			  device->devstat.devno);
		return tape_3590_erp_basic(device, request, -ENOSPC);
	case 0x3020:
		DBF_EVENT(2, "(%x): End of Data Mark\n",
			  device->devstat.devno);
		return tape_3590_erp_basic(device, request, -ENOSPC);
	case 0x3122:
		DBF_EVENT(2, "(%x): Rewind Unload initiated\n",
			  device->devstat.devno);
		return tape_3590_erp_basic(device, request, -EIO);
	case 0x3123:
		DBF_EVENT(2, "(%x): Rewind Unload complete\n",
			  device->devstat.devno);
		tape_med_state_set(device, MS_UNLOADED);
		return tape_3590_erp_basic(device, request, 0);
	case 0x4010:
		PRINT_WARN("(%x): Tape operation when medium not loaded\n",
			   device->devstat.devno);
		tape_med_state_set(device, MS_UNLOADED);
		return tape_3590_erp_basic(device, request, -ENOMEDIUM);
	case 0x4012: /* Device Long Busy */
		tape_3590_print_era_msg(device);
		return tape_3590_erp_long_busy(device, request);
	case 0x5010:
		if (sense->rac == 0xd0) {
			tape_3590_print_era_msg(device);
			return tape_3590_erp_swap(device, request);
		}
		if (sense->rac == 0x26) {
			tape_3590_print_era_msg(device);
			return tape_3590_erp_read_opposite(device, request);
		}
		return tape_3590_erp_basic(device, request, -EIO);
	case 0x5020:
	case 0x5021:
	case 0x5022:
	case 0x5040:
	case 0x5041:
	case 0x5042:
		tape_3590_print_era_msg(device);
		return tape_3590_erp_swap(device, request);
	case 0x5110:
	case 0x5111:
		return tape_3590_erp_basic(device, request, -EMEDIUMTYPE);
	case 0x5120:
	case 0x1120:
		tape_med_state_set(device,MS_UNLOADED);
		return tape_3590_erp_basic(device, request, -ENOMEDIUM);
	case 0x6020:
		PRINT_WARN("(%x): Cartridge of wrong type ?\n",
			   device->devstat.devno);
		return tape_3590_erp_basic(device, request, -EMEDIUMTYPE);
	case 0x8011:
		PRINT_WARN("(%x): Another host has reserved the tape device\n",
			   device->devstat.devno);
		TAPE_SET_STATE(device, TAPE_STATUS_BOXED);
		return tape_3590_erp_basic(device, request, -EPERM);
	case 0x8013:
		PRINT_WARN("(%x): Another host has priviliged access to the "
			   "tape device\n", device->devstat.devno);
		PRINT_WARN("(%x): To solve the problem unload the current "
			   "cartridge!\n", device->devstat.devno);
		return tape_3590_erp_basic(device, request, -EPERM);
	default:
		return tape_3590_erp_basic(device, request, -EIO);
	}
}

/*
 * 3590 interrupt handler:
 */
static int
tape_3590_irq(struct tape_device *device, struct tape_request *request)
{
	if (request == NULL)
		return tape_3590_unsolicited_irq(device);

	if ((device->devstat.dstat & DEV_STAT_UNIT_EXCEP) &&
	    (device->devstat.dstat & DEV_STAT_DEV_END) &&
	    (request->op == TO_WRI)) {
		/* Write at end of volume */
		DBF_EVENT(2, "End of volume\n");
		return tape_3590_erp_failed(device, request, -ENOSPC);
	}	

	if (device->devstat.dstat & DEV_STAT_UNIT_CHECK)
		return tape_3590_unit_check(device, request);

	if (device->devstat.dstat & DEV_STAT_DEV_END)
		return tape_3590_done(device, request);

	if (device->devstat.dstat & DEV_STAT_CHN_END) {
		DBF_EVENT(2, "cannel end\n");
		return TAPE_IO_PENDING;
	}

	if (device->devstat.dstat & DEV_STAT_ATTENTION) {
		/*
		 * We must check whether the attention message did interfere
		 * with an I/O operation. If that happened the first request
		 * on the queue has to be restarted.
		 */
		if (
			(device->devstat.flag & DEVSTAT_STATUS_PENDING) &&
			(device->devstat.flag & DEVSTAT_START_FUNCTION)
		) {
			return TAPE_IO_RETRY;
		}

		DBF_EVENT(2, "Unit Attention when busy..\n");
		return TAPE_IO_PENDING;
	}

	DBF_EVENT(6, "xunknownirq\n");
	PRINT_ERR("Unexpected interrupt.\n");
	PRINT_ERR("Current op is: %s", tape_op_verbose[request->op]);
	tape_dump_sense (device, request);
	return TAPE_IO_STOP;
}

/*
 * Setup device function
 */
static int
tape_3590_setup_device (struct tape_device  *device)
{
	struct tape_3590_disc_data *data;

	DBF_EVENT(6, "3590 device setup: %x\n", device->first_minor);
	data = kmalloc(sizeof(struct tape_3590_disc_data),
		       GFP_KERNEL | GFP_DMA);
	if (data == NULL)
		return -ENOMEM;
	data->read_back_op = READ_PREVIOUS;
	device->discdata = (void *) data;

	/* Try to find out if medium is loaded */
	if(!TAPE_BOXED(device))
		tape_3590_sense_medium(device);

	return 0;
}

/*
 * Cleanup device function
 */
void
tape_3590_cleanup_device(struct tape_device *device)
{
	if (device->discdata) {
		kfree(device->discdata);
		device->discdata = NULL;
	}
}

/*
 * List of 3590 magnetic tape commands.
 */
static tape_mtop_fn tape_3590_mtop[TAPE_NR_MTOPS] =
{
	[MTRESET]	 = tape_std_mtreset,
	[MTFSF]		 = tape_std_mtfsf,
	[MTBSF]		 = tape_std_mtbsf,
	[MTFSR]		 = tape_std_mtfsr,
	[MTBSR]		 = tape_std_mtbsr,
	[MTWEOF]	 = tape_std_mtweof,
	[MTREW]		 = tape_std_mtrew,
	[MTOFFL]	 = tape_std_mtoffl,
	[MTNOP]		 = tape_std_mtnop,
	[MTRETEN]	 = tape_std_mtreten,
	[MTBSFM]	 = tape_std_mtbsfm,
	[MTFSFM]	 = tape_std_mtfsfm,
	[MTEOM]		 = tape_std_mteom,
	[MTERASE]	 = tape_std_mterase,
	[MTRAS1]	 = NULL,
	[MTRAS2]	 = NULL,
	[MTRAS3]	 = NULL,
	[MTSETBLK]	 = tape_std_mtsetblk,
	[MTSETDENSITY]	 = NULL,
	[MTSEEK]	 = tape_3590_mtseek,
	[MTTELL]	 = tape_3590_mttell,
	[MTSETDRVBUFFER] = NULL,
	[MTFSS]		 = NULL,
	[MTBSS]		 = NULL,
	[MTWSM]		 = NULL,
	[MTLOCK]	 = NULL,
	[MTUNLOCK]	 = NULL,
	[MTLOAD]	 = tape_std_mtload,
	[MTUNLOAD]	 = tape_std_mtunload,
	[MTCOMPRESSION]	 = tape_std_mtcompression,
	[MTSETPART]	 = NULL,
	[MTMKPART]	 = NULL
};

/*
 * Tape discipline structure for 3590.
 */
static struct tape_discipline tape_discipline_3590 = {
	.owner = THIS_MODULE,
	.cu_type = 0x3590,
	.setup_device = tape_3590_setup_device,
	.cleanup_device = tape_3590_cleanup_device,
	.process_eov = tape_std_process_eov,
	.irq = tape_3590_irq,
	.read_block = tape_std_read_block,
	.write_block = tape_std_write_block,
	.assign = tape_std_assign,
	.unassign = tape_std_unassign,
#ifdef TAPE390_FORCE_UNASSIGN
	.force_unassign = tape_std_force_unassign,
#endif
#ifdef CONFIG_S390_TAPE_BLOCK
	.bread = tape_3590_bread,
	.free_bread = tape_3590_free_bread,
	.check_locate = tape_3590_check_locate,
#endif
	.ioctl_fn = tape_3590_ioctl,
	.mtop_array = tape_3590_mtop
};

/*
 * Setup discipline structure.
 */
int
tape_3590_init (void)
{
	int rc;

	TAPE_DBF_AREA = debug_register ( "tape_3590", 1, 2, 4*sizeof(long));
	debug_register_view(TAPE_DBF_AREA, &debug_sprintf_view);

	DBF_EVENT(3, "3590 init: $Revision: 1.9.4.5 $\n");
	rc = tape_register_discipline(&tape_discipline_3590);
	if (rc)
		DBF_EVENT(3, "3590 init failed\n");
	else
		DBF_EVENT(3, "3590 registered\n");
	return rc;
}

void
tape_3590_exit(void)
{
	tape_unregister_discipline(&tape_discipline_3590);
	debug_unregister(TAPE_DBF_AREA);
}

MODULE_AUTHOR("(C) 2001,2006 IBM Corporation");
MODULE_DESCRIPTION("Linux on zSeries channel attached 3590 tape device driver");
MODULE_LICENSE("GPL");

module_init(tape_3590_init);
module_exit(tape_3590_exit);
